"""
Module for generating, checking, and interpreting BIP39_ mnemonics

.. _BIP39: https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki
"""

import hashlib
import os
from unicodedata import normalize as _normalize

from .util import convertbits

class _WordList(tuple):
    """
    A tuple word list with a compact repr. Useable as a full language word-list
    tuple but doesn't spam the screen when printed in documentation. This must
    be loaded from a text file.
    """

    def __new__(cls, fn):
        """
        Create a WordList from a text file. This is a tuple of each word in the
        file
        """
        fullpath = os.path.join(os.path.dirname(os.path.realpath(__file__)), fn)
        with open(fullpath, "r") as fd:
            return super().__new__(cls, fd.read().split())

    def __init__(self, fn):
        self.filename = fn

    def __repr__(self):
        return self.__class__.__name__ + f"(\"{self.filename}\")"

#: Languages supported by this module
LANGS = ["english", "japanese", "french", "italian", "korean", "spanish"]
WORDLISTS = {lang : _WordList(f"wordlists/{lang}.txt") for lang in LANGS}

class Mnemonic(tuple):
    """
    A BIP39_ Mnemonic: a tuple of :math:`3N` spoken-language words for carrying
    seed entropy for generating deterministic wallets. The word sequence itself
    encodes a checksum for error detection. :meth:`~Mnemonic.to_seed` provides
    the interface to extract the entropy bytes, optionally decrypted with
    password, from a Mnemonic. These bytes comprise the entropy carried by the
    :class:`Mnemonic` instance and are either 128, 160, 192, 224, or 256 bits
    long depending on the length of the mnemonic.

    :class:`Mnemonic` instances can be created directly from an iterable of
    strings, a single string of whitespace-delimited words, an existing entropy
    source, or generated randomly using system entropy sources. When creating a
    new :class:`Mnemonic` instance from an existing string or iterable of
    strings, the source data must be consistent with the BIP39 Mnemonic
    validity requirements, namely that the there must be 12, 15, 18, 21, or 24
    words, each word must be a member of the wordlist specified by `lang`, and
    the checksum must pass check.  All instances of :class:`Mnemonic` are
    automatically validity-checked at construction time for these validity
    critera making it difficult to have an invalid :class:`Mnemonic` instance.
    If :class:`Mnemonic` is called with invalid initialization data, an
    Exception with a descriptive error message is raised.

    args:
        data (flexible): Data to use to initialize the mnemonic.
                         Interpreted according to its type. See below for
                         details.
        lang (str): name of spoken language to use for wordlist


    The `data` argument controls the initialization of the mnemonic data.
    It can be a number of different types, and the type of the argument
    passed in determines the initialization behavior:

    - :class:`int`:   generate random `data`-byte mnemonic (default=20)
    - :class:`bytes`: generate mnemonic using `data` bytes directly as entropy source
    - :class:`str`:   interpret as an already-made space-delimited mnemonic phrase
    - iterable:       interpret as iterable of mnemonic words


    Attributes:
        language (str): name of the language of the wordlist.

    Examples:

    Construct a mnemonic from cryptographically secure random entropy source

    >>> Mnemonic()

    Control the entropy of mnemonic by passing :class:`int` values to the
    `data` argument

    >>> Mnemonic(12)
    >>> Mnemonic(32)

    Construct a mnemonic from a predifed list of words

    >>> words = ('crunch', 'subway', 'patch', 'thrive', 'advice', 'math',
    >>>          'another', 'solid', 'horror', 'wedding', 'dance', 'smart')
    >>> Mnemonic (words)
    >>> # It works with space delimited strings too
    >>> Mnemonic ("crunch subway patch thrive advice math another solid horror"
    >>>           "wedding dance smart")

    Construct a Mnemonic from some given entropy

    """

    def __new__(cls, data=20, lang="english"):
        if isinstance(data, int):
            entropy = os.urandom(data)
            return cls._from_entropy(entropy, lang)
        if isinstance(data, bytes):
            return cls._from_entropy(data, lang)
        if isinstance(data, str):
            return cls(_normalize("NFKD", data).split(), lang)
        return super().__new__(cls, data)

    def __init__(self, data=20, lang="english"):
        self.language = lang
        self.wordlist = WORDLISTS[lang]
        self._check()

    def __str__(self):
        return ' '.join(self)

    def to_seed(self, password: bytes = b'') -> bytes:
        r"""
        Return BIP39 standard seed bytes carrying the entropy of the mnemonic,
        optionally decrypted with a password `password`.  See `Algorithm
        Description Spec`_ for details. Always returns 512-bit (64 byte) seed
        regardless of actual entropy content.

        .. _Algorithm Description Spec:  https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki#from-mnemonic-to-seed

        Examples:

        >>> source_entropy = b'\x1d\x12\xbf\xae"\xcc&w\xa2I\xab>\x8d\n\xdfk'
        >>> m = Mnemonic(source_entropy)
        >>> m
        ('brown', 'nominee', 'twist', 'easily', 'second', 'design', 'matrix', 'cube', 'direct', 'hair', 'result', 'stone')

        >>> # The seed is not equal to the initializtion entropy
        >>> m.to_seed()
        b'\r\x9a\xf0\xccq5\xd7>\xf3\xb70 \x7f\xc8\x93M\xcb\x1en+\r\xd4U\xb2\r\x19-\xaa\x99\x93\x17\xfeS\x11\xb6\nB"?\x96\x04\x7f\x8d\xeeC\x06\xcd\xdb\x9fb9\x91\x12C\xf0\xb7\x04[r\xa4H\xef\xeew'

        >>> # seed can be decrypted with a password.
        >>> m.to_seed(b"some_password")
        b'?\xb8\x18\x8a\xd5meE\x89\x08~\xfd_d\xcc6\x00\xe0\xb2(\xab\xd4\x1a\xb6\xb9ky\x93>\xe1\x9a~\x05e\xb0\xc1(R\xd6\xf6|\xab\x1a\x06\xd9j\xf7Wp,\xe5>\xa5$\xc6\xb6\xc5\x1f\xf3k\xa2F\x8a\xcb'
        """
        mnemonic_bytes = str(self).encode()
        return hashlib.pbkdf2_hmac(hash_name='sha512',
                                   password=mnemonic_bytes,
                                   salt=b'mnemonic' + password,
                                   iterations=2048)

    def _bin_string(self) -> str:
        'return str of binary representation'
        return ''.join(bin(self.wordlist.index(x))[2:].zfill(11) for x in self)

    @classmethod
    def _from_entropy(cls, ent, lang):
        """
        Create a Mnemonic from entropy bytes using given wordlist (default
        English)
        """

        wl = WORDLISTS[lang]
        ENT = len(ent) * 8
        assert len(ent) % 4 == 0 and len(ent) >= 16 and len(ent) <= 32,\
            'entropy length must be integer multiple of 32 between 128-256'

        hash_ = hashlib.sha256(ent).digest()
        chk = convertbits(hash_, 8, 1)[:ENT // 32]
        entbits = convertbits(ent, 8, 1)
        full = entbits + chk
        l = convertbits(full, 1, 11)
        return cls(tuple(wl[x] for x in l), lang)

    def _check(self):
        """
        Check if a Mnemonic instance is valid. Checks checksum and other
        validity criteria specified by BIP39. Raises descriptive exceptions on
        failure.
        """
        if len(self) not in [12, 15, 18, 21, 24]:
            raise ValueError("Incorrect Mnemonic length. must be 12, 15, 18, "
                             "21, or 24 words"
            )

        if not all(x in self.wordlist for x in self):
            baditems = frozenset(self) - frozenset(self.wordlist)
            raise ValueError(
                "Bad word(s): the following words are not "
                "present in the wordlist: " f"{list(baditems)}"
            )

        l = [self.wordlist.index(x) for x in self]
        fullbits = convertbits(l, 11, 1)
        ENT = 32 * len(fullbits) // 33
        plbits, csbits = fullbits[:ENT], fullbits[ENT:]
        plbytes = bytes(convertbits(plbits, 1, 8))
        hash_ = hashlib.sha256(plbytes).digest()
        if not convertbits(hash_, 8, 1)[:len(csbits)] == csbits:
            raise ValueError("Bad mnemonic checksum")



"Convert bitcoin addresses between legacy and cashaddr format"

import argparse
from based58 import b58encode_check, b58decode_check

from .cashaddr import cashenc, cashdec, is_cashaddr

_b58checkenc = lambda x: b58encode_check(x).decode()
_b58checkdec = lambda x: b58decode_check(x.encode())


def convert_word(word):
    """
    Parse and analyze a string as an address and return version byte info and
    both legacy and cashaddr formatted addresses corresponding to the hash
    """
    intype, decfun, encfun, p2shvbyte = (
        ("CASHAD", cashdec, _b58checkenc, b"\5") if is_cashaddr(word) else
        ("LEGACY", _b58checkdec, cashenc, b"\10")
    )
    pl = decfun(word)
    ivbyte, _hash = pl[:1], pl[1:]
    ovbyte = b"\0" if ivbyte == b"\0" else p2shvbyte
    outaddr = encfun(ovbyte + _hash)
    legaddr, cashaddr = (
        (word, outaddr) if intype == "LEGACY" else (outaddr, word)
    )
    return ivbyte, ovbyte, intype, legaddr, cashaddr

def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("file", nargs="?", default="-",
                        type=argparse.FileType('r'))
    args = parser.parse_args()
    txt = args.file.read()

    for lineno, line in enumerate(txt.split("\n")):
        for wordno, word in enumerate(line.split()):
            try:
                ivbyte, ovbyte, intype, legaddr, cashaddr = convert_word(word)
            except Exception as e:
                print(f"{'':14}ERROR  {word} {type(e).__name__}: {e}")
                continue
            print(
                f"{lineno:4d} {wordno:2d} "
                f"{ivbyte.hex().upper():2} {ovbyte.hex().upper():2} "
                f"{intype:<6} {legaddr:<34} {cashaddr}"
            )

if __name__ == "__main__":
    main()



"""
BIP32 library module. Provides classes for representing BIP32 key tree nodes.
Each node encapsulates key data and implements key derivation methods used in
BIP32
"""

import hashlib
import hmac
from functools import lru_cache
from collections import namedtuple
from os import urandom

from based58 import b58encode_check, b58decode_check

from .ecc import Point, G, N
from .cashaddr import cashenc

def hash160(msg: bytes) -> bytes:
    """
    Compute standard HASH160 of message bytes. This is the RIPEMD160 hash of
    the SHA256 hash of they message bytes.

    args:
        msg: message bytes to hash

    returns:
        bytes
    """
    shadigest = hashlib.new('sha256', msg).digest()
    return hashlib.new('ripemd160', shadigest).digest()

def node_from_str(s: str):
    """
    Create and return a BIP32 Node from a BIP32 xkey string. This takes an
    string encoded according to the `BIP 32 Serialization Format`_ and returns
    the deserialized BIP32Node instance.  Detection of key type is automatic:
    ``xprv`` strings return :class:`PrivBIP32Node` and xpub strings return
    :class:`PubBIP32Node`. Only Bitcoin style (xpub and xpriv) formats are
    supported.

    .. _BIP 32 Serialization Format: https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki#Serialization_format
    """
    b = b58decode_check(s.encode())
    c, Kbytes = b[-65:-33], b[-33:]
    depth, fingerp, index = b[4], b[5:9], int.from_bytes(b[9:13], 'big')
    if b[:4] == PrivBIP32Node.vbytes:
        assert Kbytes[0] == 0
        k = int.from_bytes(Kbytes, 'big')
        return PrivBIP32Node(k, c, depth, fingerp, index)
    if b[:4] == PubBIP32Node.vbytes:
        K = Point.from_bytes(Kbytes)
        return PubBIP32Node(K, c, depth, fingerp, index)
    raise ValueError("bad BIP32 node encoding")

class ProtocolError(ValueError):
    pass

class XPubKey(namedtuple("XKey", ["keydata", "chaincode"])):

    """
    Representation of an extended public key. This consists of an ECDSA public
    key and a 256 bit chaincode. The primary use of the :class:`XPubKey` is key
    derivation. Each XPubKey can deterministically derive :math:`2^{31}`
    non-hardened child XPubKeys each with their own unique chaincode and ECDSA
    public key making them useful for building key trees. It is not possible to
    derive any hardened child XPubKeys or any :class:`XPrivKey` s from an
    :class:`XPubKey`.
    """
    __slots__ = ()

    def __new__(cls, k, cc=None):
        cc = urandom(32) if cc is None else cc
        return super().__new__(cls, k, cc)

    @property
    def pubkey(self):
        'Public Key Curve Point (ecc.Point)'
        return self.keydata

    def addr(self, vbyte: bytes = b'\0') -> str:
        """
        Bitcoin P2PKH address with version byte `vbyte`. A ``str`` is returned
        using base58check encoding
        """
        return b58encode_check(vbyte + self.id).decode()

    def cashaddr(self):
        "Bitcoin Cash cashaddr string"
        return cashenc(b"\0" + self.id)

    @property
    def id(self):
        'Key ID. HASH160 of compressed serialized PubKey (bytes)'
        return hash160(bytes(self.pubkey))

    def __bytes__(self):
        "SEC1 compressed-form byte encoding of the ECDSA pubkey (bytes)"
        return bytes(self.pubkey)


    def ckd(self, i: int) -> "XPubKey":
        """
        Derive and return the `i`:superscript:`th` indexed child
        :class:`XPubKey`. Since children with index higher than ``0x80000000``
        are hardened, `i` must be less than ``0x80000000`` because XPubs can't
        derive hardened children.
        """
        if not i < 0x80000000:
            raise ProtocolError(
                "It is disallowed to derive a hardend subkey from public node"
            )
        pl = bytes(self.pubkey) + i.to_bytes(4, 'big')
        I = hmac.new(self.chaincode, pl, 'sha512').digest()
        IL, IR = I[:32], I[-32:]
        k = int.from_bytes(IL, 'big')
        return XPubKey(G * k  + self.keydata, IR)

    def derive(self, path):
        """
        Given a string, `path`, traverse the key tree deriving each subsequent
        node. Path is specified as strings using POSIX-like format:

        .. code-block::

            a[H]/b[H]/c[H]/d[H]/...

        where a,b,c,d.... are positive integers each optionally suffixed with
        'H'. The integers are the child indices at each level and the 'H'
        signifies that a node is hardened.
        """
        if not path:
            return self
        key = self
        for x in path.split('/'):
            x = int(x) if x[-1] != 'H' else int(x[:-1]) + 0x80000000
            key = key.ckd(int(x))
        return key

    def __str__(self):
        cc = self.chaincode.hex().upper()
        keydat = bytes(self).hex().upper()
        return f"chaincode: {cc}\nkeydata  : {keydat}"

class XPrivKey(XPubKey):
    """
    Similar to XPubKey, except that the encapsulated key is an ECDSA private
    key. This extends XPubKey with private key only capabilities such as
    exposing WIF. Also, unlike XPubKey, XPrivKey **is** capable of deriving
    hardened children (so it can derive all :math:`2^{32}` child keys)

    .. warning::
        If this Key is derived from a non-hardened parent node, and the
        non-hardened parent node's chaincode is known, then leaking this node's
        private key also leaks the parent's private key and therefore the
        entire key tree rooted there.
    """

    __slots__ = ()

    @classmethod
    def from_entropy(cls, seed):
        """
        Create and return a XprivKey from entropy bytes using the BIP32
        standard private key derivation protocol.
        """
        I = hmac.new(b'Bitcoin seed', seed, 'sha512').digest()
        k, c = int.from_bytes(I[:32], 'big'), I[32:]
        assert k and k < N, 'Invalid privkey, use  different entropy'
        return cls(k, c)

    def ckd(self, i: int):
        """
        Derive and return the `i`:superscript:`th` child XPrivKey.

        args:
            i: child key index

        returns:
            A new XPrivKey
        Note:
            indices over ``0x80000000`` are hardened children.
        """
        plbe = (XPrivKey.__bytes__(self) if i >= 0x80000000
                else bytes(self.pubkey))
        ibytes = i.to_bytes(4, 'big')
        I = hmac.new(self.chaincode, plbe + ibytes, 'sha512').digest()
        IL, IR = I[:32], I[-32:]
        k = (int.from_bytes(IL, 'big') + self.keydata) % N
        return XPrivKey(k, IR)

    def wif(self, vbyte=b'\x80'):
        'WIF string privkey'
        return b58encode_check(vbyte + self.keydata.to_bytes(32, 'big') + b'\x01').decode()

    @property
    @lru_cache()
    def pubkey(self):
        return G * self.keydata

    def __bytes__(self):
        """
        The byte 0x00 followed by the 32-byte big endian bytes representation
        of the private key
        """
        return b'\0' + self.keydata.to_bytes(32, 'big')

_pubnodebase = namedtuple("PubBIP32",
    "keydata, chaincode, depth, parent_fingerprint, index")
class PubBIP32Node(_pubnodebase, XPubKey):
    """
    Same as a XPubKey but it also tracks some additional tree position data
    during key derivation and implements BIP32-standardized serialization
    format.
    """

    __slots__ = ()
    vbytes = b'\x04\x88\xB2\x1E'    # bitcoin xpubkey version bytes in BIP32

    def __new__(cls, kd, cc, depth=0, parent_fingerprint=b'\0' * 4, index=0):
        """
        Initialize a new PubBIP32Node with

        Parameters
        ----------
        keydata : pubkey
        depth : int, optional
            tree depth of this node from the master tree root.
        parent_fingerprint : bytes
            first four bytes of the parent nodes id (address)
        index : int, optional
            child index
        """
        return super().__new__(cls, kd, cc, depth, parent_fingerprint, index)

    def __bytes__(self):
        " Return the bytes of the BIP32 extended key serialization."
        depth = self.depth.to_bytes(1, 'big')
        fingr = self.parent_fingerprint
        chnum = self.index.to_bytes(4, 'big')
        ccode = self.chaincode
        keydt = super().__bytes__()
        return  self.vbytes + depth + fingr + chnum + ccode + keydt

    def __str__(self):
        fingr = self.parent_fingerprint.hex().upper()
        cc = self.chaincode.hex().upper()
        keydat = super().__bytes__().hex().upper()
        return (f"depth    : {self.depth:d}\nindex    : {self.index:08X}\n"
                f"parent   : {fingr}\nchaincode: {cc}\nkeydata  : {keydat}\n"
                f"BIP32 str: {self.xpub}")

    @property
    def xpub(self):
        "BIP32 xpub string encoding"
        return b58encode_check(bytes(self)).decode()

    def ckd(self, i):
        xkey = super().ckd(i)
        depth = self.depth + 1
        finger = self.id[:4]
        return type(self)(*xkey, depth, finger, i)

class PrivBIP32Node(PubBIP32Node, XPrivKey):
    """
    Same as a XPrivKey but it also tracks some additional tree position data
    during key derivation and implements BIP32-standardized serialization
    format.

    .. inheritance-diagram:: PrivBIP32Node
       :parts: 1
    """
    __slots__ = ()
    vbytes = b'\x04\x88\xAD\xE4'

    @property
    def xpub(self):
        return PubBIP32Node(self.pubkey, *self[1:]).xpub

    @property
    def xprv(self):
        return b58encode_check(bytes(self)).decode()

    def __str__(self):
        return super().__str__() + f"\n         : {self.xprv}"



"""
This module provides a set of functions for implementing the cashaddr_ codec, a
codec used to format Bitcoin Cash Addresses.

The main purpose of this module is to provide cashaddr encoding and decoding
via the :func:`cashenc` and :func:`cashdec` functions. These functions
implement the full cashaddr codec. They can encode arbitrary binary data
payloads (represented as :class:`bytes`) to cashaddr strings and decode
cashaddr strings back to the original raw :class:`bytes` payload, each with
support for arbitrary user-defined cashaddr prefixes_ and checksum_
generation/validation.

The :func:`cashdec` and :func:`cashenc` functions completely provide the codec,
and while these two functions alone can probably support most useful
application development, all the other "ingredients" are also exposed by this
module to support other implementations of the codec if it is desired.

About the Codec
...............

The cashaddr codec is a base-32 ASCII-encoded format for arbitrary binary data,
with some specific design features that make it particularly suitable for
encoding cryptocurrency addresses in a secure way. The format encodes the
payload using a 32-element alphabet (:data:`ALPHABET`) designed for easy human
readability and is is completely like-cased to aid spoken communication of hash
digests by supporting case insensitive decoding. It also includes a `BCH
code`_ checksum and an arbitrary user-defined prefix which is under the
checksum to ensure proper application use during checksum validation.

The cashaddr string consists of 2 parts separated by a colon ``:``. The first
part is the human-readable prefix which can be any arbitrary string chosen by
the user. The second part is the payload and checksum_ which are encoded using
the 32-element cashaddr alphabet. This part encodes the payload data and has a
BCH checksum which is computed over both the payload data and the
human-readable prefix to ensure proper application use of the cashaddr string.

.. _cashaddr: https://github.com/bitcoincashorg/bitcoincash.org/blob/master/spec/cashaddr.md
.. _BCH code: https://en.wikipedia.org/wiki/BCH_code
.. _prefixes: https://github.com/bitcoincashorg/bitcoincash.org/blob/master/spec/cashaddr.md#prefix
.. _checksum: https://github.com/bitcoincashorg/bitcoincash.org/blob/master/spec/cashaddr.md#checksum
"""
import re

from .util import convertbits

#: cashaddr alphabet
ALPHABET = 'qpzry9x8gf2tvdw0s3jn54khce6mua7l'

_GEN = [0x98F2BC8E61, 0x79B76D99E2, 0xF33E5FB3C4, 0xAE2EABE2A8, 0x1E4F43E470]


def polymod(data: bytes) -> int:
    "Return the polymod of input byte sequence `data` over :math:`GF(2^5)`"
    c = 1
    for d in data:
        c0 = c >> 35
        c = ((c & 0x07FFFFFFFF) << 5) ^ d
        for i in range(5):
            c ^= _GEN[i] if ((c0 >> i) & 1) else 0
    return c ^ 1

def b32decode(l: str) -> list:
    """
    Decode base32-encoded string `l` into list of integers indicating the
    indices into the alphabet of the corresponding characters. Equivalent to

    .. code-block:: python

        [ALPHABET.find(x) for x in l]

    Mainly used as helper to :func:`cashdec`
    """
    return [ALPHABET.find(x) for x in l]

def b32encode(data: bytes) -> str:
    r"""
    Encode `data` as base32-encoded :class:`str` using the cashaddr alphabet.
    Returns a :class:`str` of the same length as `data` such that

    .. code-block:: python

        output[i] == ALPHABET[input[i]]

    All bytes in `data` must be less than 32. Mainly used as helper to
    :func:`cashenc`

    **Examples:**

    Encode some random bytes

    >>> payload_bytes = b"\x0f\x02\x18\x02\n\x13\r\x02\x0f\x01"
    >>> # [15, 2, 24, 2, 10, 19, 13, 2, 15,  1]
    >>> b32encode(payload_bytes)
    '0zcz2ndz0p'

    It also works on the :class:`list`-of-:class:`int` form

    >>> payload_int_list = list(payload_bytes)
    >>> b32encode(payload_int_list)
    '0zcz2ndz0p'
    """
    return "".join([ALPHABET[x] for x in data])

def prefix_expand(prefix: str) -> bytes:
    """
    Return a :class:`bytes` representation of the UTF-8-encoded bytes of
    `prefix` but with the leading three bits of each byte cleared and a zero
    byte appended at the end. This is used in the calculation of the `cashaddr
    checksum`_

    .. _cashaddr checksum: https://github.com/bitcoincashorg/bitcoincash.org/blob/master/spec/cashaddr.md#checksum
    """
    return bytes([ord(x) & 0x1f for x in prefix] + [0])

def calculate_checksum(prefix: str, payload: bytes) -> bytes:
    """
    With a given prefix string `prefix`, and B32 payload bytes `payload`,
    return the cashaddr checksum bytes. The B32 payload bytes are the payload
    message octets expanded with :func:`sorzun.util.convertbits` to 5-bit
    symbols represented as bytes.
    """
    poly = polymod(prefix_expand(prefix) + payload + b"\0"* 8)
    return bytes([((poly >> 5 * (7 - i)) & 0x1f) for i in range(8)])

def verify_checksum(prefix: str, payload: bytes) -> bool:
    """
    With a given prefix string, `prefix` and a payload bytes `payload`, verify
    that the checksum (at the end of the payload) is valid. Returns ``True``
    if the checksum is valid and ``False`` otherwise.
    """
    return polymod(prefix_expand(prefix) + payload) == 0

def cashenc(pl: bytes, prefix: str = "bitcoincash") -> str:
    r"""
    Return the cashaddr :class:`str` representation of the binary payload
    :class:`bytes` `pl` using human-readable prefix `prefix`.

    **Examples**:

    Encode binary payload

    >>> pl = b'\x00z#M\xdf\xf0\xaeh\x1fe\xc4\x99\x9d\xad9\x9e\xae\xa2\x92\xca\t'
    >>> cashenc(pl)
    'bitcoincash:qpazxnwl7zhxs8m9cjvemtfen6h29yk2pyucpwmjvj'

    Use a custom human-readable prefix

    >>> cashenc(pl, "myprefix")
    'myprefix:qpazxnwl7zhxs8m9cjvemtfen6h29yk2py5arywdw5'

    Both decode to the same original binary payload

    >>> cashdec('myprefix:qpazxnwl7zhxs8m9cjvemtfen6h29yk2py5arywdw5') == pl
    True
    >>> cashdec('bitcoincash:qpazxnwl7zhxs8m9cjvemtfen6h29yk2pyucpwmjvj') == pl
    True
    """
    pl32 = bytes(convertbits(pl, 8, 5))
    checksum = calculate_checksum(prefix, pl32)
    return prefix + ":" + b32encode(pl32 + checksum)

def cashdec(s: str) -> bytes:
    r"""
    Decode cashaddr encoded string and return :class:`bytes` payload. Decoding
    is case insensitive in the payload part. Raises :class:`AssertionError` if
    checksum validation fails. Raises :class:`ValueError` with message
    describing value and position of the first bad character if characters not
    in the base32 :data:`ALPHABET` are used.

    **Examples**:

    Decode a cashaddr string

    >>> cashdec("bitcoincash:qpazxnwl7zhxs8m9cjvemtfen6h29yk2pyucpwmjvj")
    b'\x00z#M\xdf\xf0\xaeh\x1fe\xc4\x99\x9d\xad9\x9e\xae\xa2\x92\xca\t'
    >>> # case insensitive in the payload part
    >>> cashdec("bitcoincash:qpAzXnwL7zHXs8m9cJVemtFEn6h29YK2pyUcpwmJvj")
    b'\x00z#M\xdf\xf0\xaeh\x1fe\xc4\x99\x9d\xad9\x9e\xae\xa2\x92\xca\t'

    Raises :class:`AssertionError` if checksum fails

    >>> # Bad Character---------------------↓
    >>> cashdec("bitcoincash:qpazxnwl7zhxs8mncjvemtfen6h29yk2pyucpwmjvj")
    Traceback (most recent call last):
      ...
    AssertionError: Bad checksum

    Raises :class:`ValueError` if charactars not in the :data:`ALPHABET` are
    used

    >>> cashdec("bitcoincash:alphabetsoup")
    Traceback (most recent call last):
      ...
    ValueError: invalid base32 symbol 'b' at position 5
    """
    prefix, pltxt = s.split(":")
    pltxt = pltxt.lower() # to support case insensitive input
    dec = b32decode(pltxt)
    if -1 in dec:
        badloc = dec.index(-1)
        raise ValueError(
            f"invalid base32 symbol '{pltxt[badloc]}' at position {badloc}"
        )
    pl32 = bytes(dec)
    assert verify_checksum(prefix, pl32), "Bad checksum"
    return bytes(convertbits(pl32[:-8], 5, 8, False))

def is_cashaddr(s: str) -> bool:
    """
    Return ``True`` if and only if string `s` is a cashaddr encoded string.
    Does not perform checksum validation or decoding, only checks syntax format
    and that all characters are taken from the cashaddr alphabet.
    """
    return bool(re.match(f"[a-zA-Z0-9]*:[{ALPHABET}]+$", s))



"Small utility module for some common functions"

def convertbits(
        data: bytes, frombits: int, tobits: int, pad: bool=True
    ) -> bytes:
    """
    Convert an iterable of non-negative integers `data` from base
    :math:`2^\mathrm{frombits}` to a list of base :math:`2^\mathrm{tobits}`
    symbols
    """
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret



import argparse
import math
from .deterministic import node_from_str, PrivBIP32Node
from .mnemonic import Mnemonic

def range_from_str(s):
    'return a arange from a string in x-y format'
    return range(*map(int, s.split('-'))) if '-' in s else range(int(s))

def main():
    parser = argparse.ArgumentParser(description='Key Utility')
    parser.add_argument('-p', '--path', help='derivation path')
    parser.add_argument('-w', '--wif', action='store_true',
                        help='print leaf private keys (compressed WIF format)')

    parser.add_argument('-l', default=range(25),
                        type=range_from_str,
                        help="""
                        Range of leaf indices to compute. Format is x-y. If
                        only one number is given x is assumed to be zero.
                        """)
    parser.add_argument('keydata', nargs='?', default=None,
                        help="""
                        Key specification. This can be a BIP32 standard xpub or
                        xprv key string, a space delimited string of mnemonic
                        words, a hex encoded BIP32 seed, a number of bytes to
                        use as entropy for generating a random mnemonic, or
                        blank which defaults to 20 bytes of entropy for
                        generating a random mnemonic
                        """)
    parser.add_argument('-f', '--format', help='address format', default='BTC',
                        choices=['BTC', 'LTC', "BCH"])
    parser.add_argument("--long-bch-format", action="store_true")
    args = parser.parse_args()

    addrpre = {'BTC' : b'\0', 'LTC' : b'0', "BCH" : b"\0"}
    wifpre = {'BTC' : b'\x80', 'LTC' : b'\xb0', "BCH" : b"\x80"}

    print('Root key info ' + '-' * 97)

    if args.keydata is None:
        m = Mnemonic()
    elif args.keydata.isdigit() and len(args.keydata) < 4:
        m = Mnemonic(int(args.keydata))
    elif ' ' in args.keydata:
        m = Mnemonic(args.keydata)
    elif args.keydata.startswith('xp'):
        r = node_from_str(args.keydata)
    elif set(args.keydata) <= set('1234567890abcdefABCDEF'):
        seed = bytes.fromhex(args.keydata)

    if 'm' in locals():
        print(f"Mnemonic : {m}")
        seed = m.to_seed()

    if 'seed' in locals():
        hexseed = seed.hex().upper()
        print(f"seed     : {hexseed[:64]}\n           {hexseed[64:]}")
        r = PrivBIP32Node.from_entropy(seed)

    print(r)

    if args.path:
        mend = r.derive(args.path) if args.path else r
        print(f"\nDerived Key info {'':-<94}")
        print(f"path     : {args.path}")
        print(mend)
    else:
        mend = r

    ll = math.ceil(math.log10(args.l.stop))     # index text width
    al = 34 if args.format != "BCH" else 42     # address text width
    kl = 52 if args.wif else 66                 # key text width
    # cashaddr abbreveation adjustment.
    ab = 12 if (args.long_bch_format and args.format == "BCH") else 0
    print(f"\n{'leaves':-<{ll + al + kl + ab + 2}}")
    for i in args.l:
        xkey = mend.ckd(i)

        if args.format != "BCH":
            addr = xkey.addr(addrpre[args.format])
        elif args.long_bch_format:
            addr = xkey.cashaddr()
        else:
            addr = xkey.cashaddr()[12:]

        keydat = (xkey.wif(wifpre[args.format]) if args.wif
                  else bytes(xkey.pubkey).hex().upper())
        print(f"{i:{ll}d} {addr:<34} {keydat}")

if __name__ == "__main__":
    main()
